#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const lean_client_js_node_1 = require("lean-client-js-node");
const node_1 = require("vscode-languageserver/node");
const vscode_languageserver_textdocument_1 = require("vscode-languageserver-textdocument");
const vscode_uri_1 = require("vscode-uri");
const connection = node_1.createConnection();
const server = new lean_client_js_node_1.Server(null);
server.error.on((err) => {
    switch (err.error) {
        case 'stderr':
            connection.console.log(err.chunk);
            break;
        case 'connect':
            connection.window.showErrorMessage(err.message);
            break;
        case 'unrelated':
            connection.tracer.log(err.message);
            break;
    }
});
const documents = new node_1.TextDocuments(vscode_languageserver_textdocument_1.TextDocument);
documents.listen(connection);
connection.onInitialize((params) => {
    const extraArgs = process.argv.indexOf('--');
    const argv = extraArgs === -1 ? [] : process.argv.slice(extraArgs + 1);
    server.transport = new lean_client_js_node_1.ProcessTransport('lean', params.rootPath, argv);
    server.connect();
    sendRoi();
    return {
        capabilities: {
            textDocumentSync: node_1.TextDocumentSyncKind.Full,
            completionProvider: {},
            definitionProvider: true,
            hoverProvider: true,
            documentSymbolProvider: true,
            workspaceSymbolProvider: true,
        },
    };
});
// The content of a text document has changed. This event is emitted
// when the text document first opened or when its content has changed.
documents.onDidChangeContent((change) => {
    server.sync(vscode_uri_1.URI.parse(change.document.uri).fsPath, change.document.getText());
});
function sendRoi() {
    server.roi('visible-files', documents.keys().map((uri) => ({ file_name: vscode_uri_1.URI.parse(uri).fsPath,
        ranges: [{ begin_line: 0, end_line: Math.pow(2, 31) }] })));
}
documents.onDidOpen(() => sendRoi());
documents.onDidClose(() => sendRoi());
// // The settings interface describe the server relevant settings part
// interface Settings {
//     languageServerExample: ExampleSettings;
// }
// // These are the example settings we defined in the client's package.json
// // file
// interface ExampleSettings {
//     maxNumberOfProblems: number;
// }
// hold the maxNumberOfProblems setting
// let maxNumberOfProblems: number;
// // The settings have changed. Is send on server activation
// // as well.
// connection.onDidChangeConfiguration((change) => {
//     const settings =  change.settings as Settings;
//     maxNumberOfProblems = settings.languageServerExample.maxNumberOfProblems || 100;
//     // Revalidate any open text documents
//     documents.all().forEach(validateTextDocument);
// });
function toSeverity(severity) {
    if (severity === 'warning') {
        return node_1.DiagnosticSeverity.Warning;
    }
    else if (severity === 'error') {
        return node_1.DiagnosticSeverity.Error;
    }
    else if (severity === 'information') {
        return node_1.DiagnosticSeverity.Information;
    }
}
server.allMessages.on((messages) => {
    const diagnosticMap = new Map();
    for (const uri of documents.keys()) {
        diagnosticMap.set(uri, []);
    }
    for (const message of messages.msgs) {
        const start_line = Math.max(message.pos_line - 1, 0);
        const end_line = Math.max(message.end_pos_line - 1, 0);
        const start_pos = node_1.Position.create(start_line, message.pos_col);
        const end_pos = message.end_pos_col !== undefined
            ? node_1.Position.create(end_line, message.end_pos_col)
            : start_pos;
        const range = node_1.Range.create(start_pos, end_pos);
        const diagnostics = diagnosticMap.get(vscode_uri_1.URI.file(message.file_name).toString());
        if (diagnostics) {
            diagnostics.push({
                range,
                message: message.text,
                severity: toSeverity(message.severity),
            });
        }
    }
    diagnosticMap.forEach((diagnostics, uri) => {
        connection.sendDiagnostics({ uri, diagnostics });
    });
});
let filesInProgress = [];
server.tasks.on((tasks) => {
    var _a;
    const newProgress = {};
    for (const task of tasks.tasks) {
        newProgress[_a = task.file_name] || (newProgress[_a] = []);
        newProgress[task.file_name].push({
            range: node_1.Range.create(node_1.Position.create(task.pos_line - 1, task.pos_col), node_1.Position.create(task.end_pos_line - 1, task.end_pos_col)),
        });
    }
    for (const oldFile of filesInProgress) {
        newProgress[oldFile] || (newProgress[oldFile] = []);
    }
    filesInProgress = [];
    for (const fileName in newProgress) {
        if (newProgress[fileName].length) {
            filesInProgress.push(fileName);
        }
        const params = {
            textDocument: { version: 0, uri: vscode_uri_1.URI.file(fileName).toString() },
            processing: newProgress[fileName],
        };
        connection.sendNotification('$/lean/fileProgress', params);
    }
});
connection.onCompletion((textDocumentPosition) => {
    const fileName = vscode_uri_1.URI.parse(textDocumentPosition.textDocument.uri).fsPath;
    const position = textDocumentPosition.position;
    return server.complete(fileName, position.line + 1, position.character).then((message) => {
        return !message.completions ? [] : message.completions.map((completion) => {
            const item = {
                label: completion.text,
                kind: toCompletionItemKind(completion.kind),
                documentation: completion.doc,
                detail: completion.tactic_params ?
                    completion.tactic_params.join(' ') :
                    completion.type,
            };
            // item.range = new vscode.Range(position.translate(0, -message.prefix.length), position);
            if (completion.tactic_params) {
                item.detail = completion.tactic_params.join(' ');
            }
            else {
                item.detail = completion.type;
            }
            item.documentation = completion.doc;
            return item;
        });
    });
});
connection.onDefinition(((params) => {
    const fileName = vscode_uri_1.URI.parse(params.textDocument.uri).fsPath;
    const position = params.position;
    return server.info(fileName, position.line + 1, position.character).then((response) => {
        if (response.record && response.record.source) {
            const src = response.record.source;
            const uri = src.file ? vscode_uri_1.URI.file(src.file).toString() : params.textDocument.uri;
            const pos = node_1.Position.create(src.line - 1, src.column);
            return [node_1.Location.create(uri, node_1.Range.create(pos, pos))];
        }
        else {
            return [];
        }
    });
}));
connection.onHover((params) => {
    const fileName = vscode_uri_1.URI.parse(params.textDocument.uri).fsPath;
    const position = params.position;
    return server.info(fileName, position.line + 1, position.character).then((response) => {
        const record = response.record;
        if (!record) {
            return { contents: [] };
        }
        const marked = [];
        const name = record['full-id'] || record.text;
        if (name) {
            if (response.record.tactic_params) {
                marked.push({
                    language: 'text',
                    value: name + ' ' + record.tactic_params.join(' '),
                });
            }
            else {
                marked.push({
                    language: 'lean',
                    value: name + ' : ' + record.type,
                });
            }
        }
        if (response.record.doc) {
            marked.push(response.record.doc);
        }
        return {
            contents: marked,
            range: node_1.Range.create(position, position),
        };
    });
});
function toCompletionItemKind(s) {
    switch (s) {
        case undefined: return node_1.CompletionItemKind.Function; // for older lean servers
        case 'meta': return node_1.CompletionItemKind.Event;
        case 'class': return node_1.CompletionItemKind.Interface;
        case 'definition': return node_1.CompletionItemKind.Field;
        case 'inductive': return node_1.CompletionItemKind.Enum;
        case 'instance': return node_1.CompletionItemKind.Constant;
        case 'structure': return node_1.CompletionItemKind.Class;
        case 'theorem': return node_1.CompletionItemKind.Function;
        default: return node_1.CompletionItemKind.Value;
    }
}
function toSymbolKind(s) {
    switch (s) {
        case undefined: return node_1.SymbolKind.Function; // for older lean servers
        case 'meta': return node_1.SymbolKind.Event;
        case 'class': return node_1.SymbolKind.Interface;
        case 'definition': return node_1.SymbolKind.Field;
        case 'inductive': return node_1.SymbolKind.Enum;
        case 'instance': return node_1.SymbolKind.Constant;
        case 'structure': return node_1.SymbolKind.Class;
        case 'theorem': return node_1.SymbolKind.Function;
        default: return node_1.SymbolKind.Object;
    }
}
connection.onDocumentSymbol((params) => __awaiter(void 0, void 0, void 0, function* () {
    // See https://github.com/leanprover/vscode-lean/blob/60c43bba6943879ecc198acb69d27f1b60c9d322/src/symbol.ts#L15
    const fileName = vscode_uri_1.URI.parse(params.textDocument.uri).fsPath;
    const response = yield server.symbols(fileName);
    const infos = [];
    let stack = [];
    for (const item of response.results) {
        if (!(item.source && item.source.file && item.source.line && item.source.column))
            continue;
        const range = node_1.Range.create(node_1.Position.create(item.source.line - 1, item.source.column), node_1.Position.create(item.source.line - 1, item.source.column + 1));
        // Pop as much stack as necessary
        let i = 0;
        for (; i < item.name_parts.length && i < stack.length; i++) {
            if (stack[i].name !== item.name_parts[i]) {
                break;
            }
        }
        stack = stack.slice(0, i);
        for (; i < item.name_parts.length; i++) {
            const curr = {
                name: item.name_parts[i],
                detail: null,
                kind: node_1.SymbolKind.Namespace,
                range: range,
                selectionRange: range,
                children: [],
            };
            if (i > 0) {
                stack[i - 1].children.push(curr);
            }
            else {
                infos.push(curr);
            }
            stack.push(curr);
        }
        stack[stack.length - 1].kind = toSymbolKind(item.kind);
        stack[stack.length - 1].detail = item.kind;
    }
    return infos;
}));
connection.onWorkspaceSymbol((params) => __awaiter(void 0, void 0, void 0, function* () {
    const response = yield server.search(params.query);
    return response.results
        .filter((item) => item.source && item.source.file &&
        item.source.line && item.source.column)
        .map((item) => {
        const loc = {
            uri: vscode_uri_1.URI.file(item.source.file).toString(),
            range: node_1.Range.create(node_1.Position.create(item.source.line - 1, item.source.column), node_1.Position.create(item.source.line - 1, item.source.column + 1))
        };
        return {
            name: item.text,
            kind: toSymbolKind(item.kind),
            type: item.type,
            location: loc
        };
    });
}));
function upconvertToLean4PlainGoal(goalState) {
    // strip 'N goals' from the front (which is present for 0 or 2+ goals)
    if (goalState === 'no goals') {
        return { rendered: '', goals: [] };
    }
    ;
    const withoutCount = goalState.replace(/^\d+ goals?\n/, '');
    const goals = withoutCount.split(/(?<=^⊢ [^]*?\n)\n/m);
    return { rendered: '', goals: goals.map(each => each.trim()) };
}
connection.onRequest('$/lean/plainGoal', (params) => __awaiter(void 0, void 0, void 0, function* () {
    const fileName = vscode_uri_1.URI.parse(params.textDocument.uri).fsPath;
    const position = params.position;
    return server.info(fileName, position.line + 1, position.character).then((response) => {
        const record = response.record;
        if (!record || !record.state) {
            return {};
        }
        return upconvertToLean4PlainGoal(record.state.trim());
    });
}));
connection.onRequest('$/lean/discoverWidget', (params) => __awaiter(void 0, void 0, void 0, function* () {
    const fileName = vscode_uri_1.URI.parse(params.textDocument.uri).fsPath;
    const position = params.position;
    const info = yield server.info(fileName, position.line + 1, position.character);
    if (!info.record || !info.record.widget) {
        return {};
    }
    const widget_info = info.record.widget;
    const widget = yield server.send({
        command: 'get_widget',
        file_name: fileName,
        line: widget_info.line,
        column: widget_info.column,
        id: widget_info.id,
    });
    return { widget: widget.widget };
}));
connection.onRequest('$/lean/widgetEvent', (params) => __awaiter(void 0, void 0, void 0, function* () {
    const widget = params.widget;
    const response = yield server.send({
        command: 'widget_event',
        kind: params.kind,
        handler: params.handler,
        args: params.args,
        file_name: vscode_uri_1.URI.parse(params.textDocument.uri).fsPath,
        line: widget.line,
        column: widget.column,
        id: widget.id,
    });
    return response;
}));
connection.listen();
